pragma solidity 0.8.12;

import "forge-std/Test.sol";
import "forge-std/console2.sol";

contract Getter {
    function getETHBalance(address addr) public view returns (uint256) {
        return address(addr).balance;
    }

    function isContract(address addr) public view returns (bool) {
        uint256 size;
        assembly {
            size := extcodesize(addr)
        }
        return size != 0;
    }

    function isERC20(address addr) public view returns (bool) {
        if (!isContract(addr)) return false;
        (bool s,) = addr.staticcall(abi.encodeWithSignature("decimals()"));
        return s;
    }

    function possiblePairs(
        address token0,
        address[] calldata token1s,
        address[] calldata factories
    ) public view returns (bool found, address pair, uint256 amount) {
        if (!isERC20(token0)) {
            return (false, address(0), 0);
        }
        for (uint256 i = 0; i < token1s.length; i++) {
            address token1 = token1s[i];
            address factory = factories[i];
            (bool s, bytes memory data) = factory.staticcall(
                abi.encodeWithSignature("getPair(address,address)", token0, token1)
            );
            if (s) {
                address pair = abi.decode(data, (address));
                if (pair == address(0)) return (false, address(0), 0);
                (bool s2, bytes memory data2) =
                    token1.staticcall(abi.encodeWithSignature("balanceOf(address)", pair));
                if (s2) {
                    uint256 amount = abi.decode(data2, (uint256));
                    if (amount > 0) {
                        return (true, pair, amount);
                    }
                }
            }
        }
        return (false, address(0), 0);
    }

    function possiblePairsV3(
        address token0,
        address[] calldata token1s,
        address[] calldata factories,
        uint24[] calldata fees
    ) public view returns (bool found, address pair, uint256 amount) {
        if (!isERC20(token0)) {
            return (false, address(0), 0);
        }
        for (uint256 i = 0; i < token1s.length; i++) {
            address token1 = token1s[i];
            address factory = factories[i];
            (bool s, bytes memory data) = factory.staticcall(
                abi.encodeWithSignature(
                    "getPool(address,address,uint24)", token0, token1, fees[i]
                )
            );
            if (s) {
                address pair = abi.decode(data, (address));
                if (pair == address(0)) return (false, address(0), 0);
                (bool s2, bytes memory data2) =
                    token1.staticcall(abi.encodeWithSignature("balanceOf(address)", pair));
                if (s2) {
                    uint256 amount = abi.decode(data2, (uint256));
                    if (amount > 0) {
                        return (true, pair, amount);
                    }
                }
            }
        }
        return (false, address(0), 0);
    }
}

contract POC is Test {
    Getter public getter;
    address alice = address(1);

    function setUp() public {
        vm.createSelectFork(vm.envString("ETH_RPC_URL"), 16038375);
        getter = new Getter();
    }

    function test_1() public {
        vm.deal(alice, 1000 ether);
        assertEq(getter.getETHBalance(alice), 1000 ether);
        assertEq(getter.isContract(alice), false);
        assertEq(getter.isContract(address(getter)), true);
        assertEq(getter.isERC20(address(getter)), false);
    }

    function test_2() public {
        address WETH = 0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2;
        address USDT = 0xdAC17F958D2ee523a2206206994597C13D831ec7;
        address factory = 0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f;
        assertEq(getter.isERC20(WETH), true);
        assertEq(getter.isERC20(USDT), true);
        assertEq(getter.isERC20(factory), false);
        address[] memory tokens = new address[](1);
        tokens[0] = USDT;
        address[] memory factories = new address[](1);
        factories[0] = factory;
        (bool found, address pair, uint256 amount) =
            getter.possiblePairs(WETH, tokens, factories);
        assertEq(found, true);
        assertEq(pair, 0x0d4a11d5EEaaC28EC3F61d100daF4d40471f1852);
        assertEq(amount, 14782372483603);

        tokens[0] = address(this);
        (found, pair, amount) = getter.possiblePairs(WETH, tokens, factories);
        assertEq(found, false);
        assertEq(pair, address(0));
        assertEq(amount, 0);
    }

    function test_3() public {
        address factoryV3 = 0x1F98431c8aD98523631AE4a59f267346ea31F984;
        address WETH = 0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2;
        address USDT = 0xdAC17F958D2ee523a2206206994597C13D831ec7;

        address[] memory tokens = new address[](1);
        tokens[0] = USDT;
        address[] memory factories = new address[](1);
        factories[0] = factoryV3;
        uint24[] memory fees = new uint24[](1);
        fees[0] = 3000;
        (bool found, address pair, uint256 amount) =
            getter.possiblePairsV3(WETH, tokens, factories, fees);
        assertEq(found, true);
        assertEq(pair, 0x4e68Ccd3E89f51C3074ca5072bbAC773960dFa36);
        assertEq(amount, 18375033354951);

        fees[0] = 1000;
        (found, pair, amount) = getter.possiblePairsV3(WETH, tokens, factories, fees);
        assertEq(found, false);
        assertEq(pair, address(0));
        assertEq(amount, 0);
    }
}
