## Binance Wallet Monitor
这是一个监控币安冷热钱包的小脚本程序，主要依据是币安在其网站上公布的冷热钱包在各条链上的地址进行监控。
币安公布地址的网址连接如下：
https://www.binance.com/en/blog/community/our-commitment-to-transparency-2895840147147652626

### 设计思路：
从各条链上讲对应的token数据抓下来，存放在数据库中。然后另起一个process程序来处理数据用作报警。
#### 获取数据
主要涉及到如下几条链：
BTC,ETH,BNB,BSC,TRX链，
其中，ETH和BSC是EVM兼容链，采用的是multicall的方式直接一次调用RPC拿到数据，针对原生代币也通过调用智能合约的方式一并拿到数据。
针对非EVM兼容链，
- BNB链，主要参考的是币安官网对于BNB链的API，连接如下：
https://docs.bnbchain.org/docs/beaconchain/develop/api-reference/dex-api/paths/#apiv1accountaddress
调用的API方法为：/api/v1/account/{address}
要求是1秒钟，最多调用5次请求
- BTC链，主要使用的是blockonomics的API数据接口，也需要注册API key参考连接如下：
https://www.blockonomics.co/views/api.html
调用的API方法为：https://www.blockonomics.co/api/balance
- TRX链，使用的是TRX官方的tronweb连接库，需要先注册一个API KEY，然后使用该API KEY来进行API请求，参考连接如下：https://cn.developers.tron.network/v3.7/docs/tron-web-class
请求的方法是：调用对应合约的balanceOf方法

请求得到的数据最后都会归一化到同一个数据格式：
```js
export type BinanceInfo = {
    "account": string,
    "balance": string,
    "decimals": number,
    "symbol": SYMBOL,
    "chain": CHAIN
}
```
然后在调用`types.ts`中的方法，将其映射到数据表对应的格式中。并插入数据库。所有的数据转换工作都在`types.ts`文件中完成。


### 报警逻辑
报警程序是独立的另一个程序，主要是查数据库，然后按规则触发报警。具体的报警规则是：
报警的标的是数量之和，跟价格无关。

1. 针对token，统计其在各条链上的数量之和，作为报警的数据源。
2. 对比当天的第一笔记录情况，如果下跌超过20%就触发报警，如果继续下跌超30%...90%也报警，然而反弹到20%则不报警
3. 对比所有记录的第一笔情况，或者上一次报警的记录情况，如果下跌超过20%就触发报警，如果继续下跌超30%也报警，然而反弹到20%则不报警


具体的实现非常依赖于以下sql的统计数据：`firstToday,last,curr`
```sql
-- 统计当天第一笔记录的balance之和: firstToday
SELECT sum(balance) as balance, avg(timestamp) as timestamp
FROM
     record_balance_bnb_binance_wallet_balance
WHERE
     token_address = 'BNB_BNB'
     and timestamp = (
          SELECT TIMESTAMP
          FROM
          record_balance_bnb_binance_wallet_balance
          where
          Date(now()) = Date(FROM_UNIXTIME(timestamp))
          ORDER BY TIMESTAMP
          LIMIT 1
     )
limit 1
-- 统计上一次报警的记录的balnace之和，如果没有上一次报警记录，则统计所有记录的第一记录的balance之和：last
SELECT sum(balance) as balance, avg(timestamp) as timestamp
FROM
     record_balance_bnb_binance_wallet_balance
WHERE
     token_address = 'BNB_BNB'
     and timestamp = IFNULL(
     (
          SELECT valueInt
          FROM
          global_params
          where
          keyStr = 'BNB_BNB__Last_Timestamp'
          ORDER BY timestamp DESC
          LIMIT 1
     ),(
          SELECT TIMESTAMP
          FROM
          record_balance_bnb_binance_wallet_balance
          where token_address = 'BNB_BNB'
          ORDER BY TIMESTAMP
          LIMIT 1
     ))
order by timestamp
limit 1
-- 统计当前最新的一次记录的balance之和：curr
SELECT sum(balance) as balance, avg(timestamp) as timestamp
FROM
     record_balance_bnb_binance_wallet_balance
WHERE
     token_address = 'BNB_BNB'
     and timestamp = (
          SELECT TIMESTAMP
          FROM
          record_balance_bnb_binance_wallet_balance
          ORDER BY TIMESTAMP DESC
          LIMIT 1
     )
limit 1
```
具体的计算逻辑是参考了FTT价格报警的算法设计，知道当前的状态curr，要比较的状态firstToday或者是last，然后计算出delta值，然后根据delta值，去表格中查询对应的index，并将index与global index进行比较，如果index>gloabl index,说明下跌幅度扩大，触发报警；否则，说明下跌幅度收窄，则不触发报警。

重置global index的逻辑是：
针对last比较，每报警一次last，则需要重置一次last 的global index；
针对24h，如果firstToday的UTC时间戳与当前系统中维护的firstToday UTC时间戳不一样时，需要重置一次firstToday的global index



### Debug
报警程序卡住了，原因是插入中间表的时候存在唯一键冲突，但是错误处理没做好，pm2没有及时拉起来。
更改了中间表的唯一键，调试确认了能够正常报警。
即将uniqueKey进行了修改
```sql
DROP table if EXISTS global_params;

create table
    `global_params` (
        `id` int NOT NULL AUTO_INCREMENT COMMENT 'id',
        `keyStr` varchar(255) NOT NULL,
        `valueString` varchar(255) NOT NULL DEFAULT '',
        `valueInt` BIGINT NOT NULL,
        `timestamp` int NOT NULL COMMENT '时间戳',
        `create_time` datetime NOT NULL DEFAULT CURRENT_TIMESTAMP COMMENT '创建时间',
        `update_time` datetime NOT NULL DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP COMMENT '修改时间',
        PRIMARY KEY (`id`),
        UNIQUE KEY `balance_idx` (
            `keyStr`,
            `valueInt`,
            `valueString`
        ) USING BTREE
    ) ENGINE = InnoDB DEFAULT CHARSET = utf8mb4;

```